// **********************************************************************
// PUCRS/Escola Politécnica
// COMPUTAÇÃO GRÁFICA
//
// Programa basico para criar aplicacoes 2D em OpenGL
//
// Marcio Sarroglia Pinho
// pinho@pucrs.br
// **********************************************************************

//
// Para uso no Xcode:
// Abra o menu Product -> Scheme -> Edit Scheme -> Use custom working directory
// Selecione a pasta onde você descompactou o ZIP que continha este arquivo.
//

#include <iostream>
#include <iomanip>
#include <cmath>
#include <ctime>

using namespace std;


#ifdef WIN32
#include <windows.h>
#include "gl\glut.h"
static DWORD last_idle_time;
#else
#include <sys/time.h>
static struct timeval last_idle_time;
#endif

#ifdef __APPLE__
#include <GLUT/glut.h>
#endif


#include "Ponto.h"
#include "Linha.h"

const int MAX = 5000;
bool devoTestar = false;

Linha Linhas[MAX];
Linha Veiculo;

float tx, ty, alfa;

// ***********************************************
// imprime a matriz de transformaçao da OpenGL
// ***********************************************
void imprimeMatrizGL() {

	GLfloat matriz_gl[4][4];

	glGetFloatv(GL_MODELVIEW_MATRIX,&matriz_gl[0][0]);

	for(int i = 0; i<4; i++) {
		for(int j=0; j<4; j++) {
            cout << setw(5) << matriz_gl[i][j] << " ";
		}
		cout << endl;
	}
}
// ***********************************************
//  void calcula_ponto(Ponto p, Ponto &out)
//
//  Esta função calcula as coordenadas
//  de um ponto no sistema de referência do
//  universo (SRU), ou seja, aplica as rotações,
//  escalas e translações a um ponto no sistema
//  de referência do objeto SRO.
// ***********************************************
void calcula_ponto(Ponto p, Ponto &out) {
	GLfloat ponto_novo[4];
	GLfloat matriz_gl[4][4];
	int  i;

	glGetFloatv(GL_MODELVIEW_MATRIX,&matriz_gl[0][0]);

	for(i=0; i<4; i++) {
		ponto_novo[i] = matriz_gl[0][i] * p.x +
		                matriz_gl[1][i] * p.y +
		                matriz_gl[2][i] * p.z +
		                matriz_gl[3][i];
	}
	out.x = ponto_novo[0];
	out.y = ponto_novo[1];
	out.z = ponto_novo[2];

	//imprimeMatrizGL();

}
// **********************************************************************
// int intersec2d(Ponto k, Ponto l, Ponto m, Ponto n, double &s, double &t)
// Calcula a intersecção entre dois segmentos de reta
// **********************************************************************
int intersec2d(Ponto k, Ponto l, Ponto m, Ponto n, double &s, double &t) {
    double det;
    
    det = (n.x - m.x) * (l.y - k.y)  -  (n.y - m.y) * (l.x - k.x);
    
    if (det == 0.0)
        return 0 ; // não há intersecção
    
    s = ((n.x - m.x) * (m.y - k.y) - (n.y - m.y) * (m.x - k.x))/ det ;
    t = ((l.x - k.x) * (m.y - k.y) - (l.y - k.y) * (m.x - k.x))/ det ;
    
    return 1; // há intersecção
}

// **********************************************************************
//  void init(void)
//  Inicializa os parâmetros globais de OpenGL
//
// **********************************************************************
void init(void) {
	// Define a cor do fundo da tela (AZUL)
	glClearColor(0.0f, 0.0f, 1.0f, 1.0f);

	srand(time(NULL));
	for(int i=0; i<MAX; i++)
		Linhas[i].geraLinhaCurta();

	// Define as coordenadas da linha que representa o veículo
	Veiculo.x1 = 0;
	Veiculo.y1 = 1.0;
	Veiculo.x2 = 0;
	Veiculo.y2 = -1.0;

	// Define a posição inicial da linha que representa o veículo
	tx = 5;
	ty = 5;
	alfa = 0.0;

}

// **********************************************************************
//  void reshape( int w, int h )
//  trata o redimensionamento da janela OpenGL
//
// **********************************************************************
void reshape( int w, int h ) {
	// Reset the coordinate system before modifying
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	// Define os limites lógicos da área OpenGL dentro da Janela
	glOrtho(0,10,0,10,0,1);

	// Define a área a ser ocupada pela área OpenGL dentro da Janela
	glViewport(0, 0, w, h);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

}


// **********************************************************************
//  void display( void )
//
// **********************************************************************
void display( void ) {
    static float dt;
    static float AccumTime=0;
    
#ifdef _WIN32
    DWORD time_now;
    time_now = GetTickCount();
    dt = (float) (time_now - last_idle_time) / 1000.0;
#else
    // Figure out time elapsed since last call to idle function
    struct timeval time_now;
    gettimeofday(&time_now, NULL);
    dt = (float)(time_now.tv_sec  - last_idle_time.tv_sec) +
    1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
#endif
    AccumTime +=dt;
    if (AccumTime >=3) // imprime o FPS a cada 3 segundos
    {
        cout << 1.0/dt << " FPS"<< endl;
        AccumTime = 0;
    }


	static int cont = 0;
	static clock_t antes;

	if (cont == 0)
		antes = clock();
	cont ++;

	Ponto P1, P2, temp;

	// Limpa a tela coma cor de fundo
	glClear(GL_COLOR_BUFFER_BIT);

	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();

	// Desenha a linha que representa o "veículo"
	glColor3f(1,0,0);
	glLineWidth(3);
	glPushMatrix();
	{
		glTranslatef(tx, ty, 0);
		glRotatef(alfa,0,0,1);
		//Veiculo.desenhaLinha();

		// guarda as coordenadas do primeiro ponto da linha
		temp.set(Veiculo.x1, Veiculo.y1);
		calcula_ponto(temp, P1);
		//cout << "Ponto Inicial: " ;
		//temp.imprime();
		//cout << endl;
		// ******************
		// Exercício 1:
		// ******************
		// faça o mesmo para o segundo ponto

		temp.set(Veiculo.x2, Veiculo.y2);
		calcula_ponto(temp, P2);
	}
	glPopMatrix();

	glLineWidth(1);
	glColor3f(0,1,0);
	// desenha as linhas do cenário
	for(int i=0; i<MAX; i++) {
		// ******************
		// Exercício 2:
		// ******************
		// Antes de desenhar cada linha, calcule se esta tem colisão com o veículo
		// Se houver, troque a cor da linha
		// 		if (TemColisao(P1, P2, ........)
		// 			glColor3f(1,0,0);
		//
		if (devoTestar) { // esta variável é controlada pela "tecla de espaço"
            
		}
		Linhas[i].desenhaLinha();
	}
	
	// desenha a linha do veículo de novo, por cima de todo o resto.
	glColor3f(1,0,1);
	glLineWidth(3);
	glPushMatrix();
	{
		glTranslatef(tx, ty, 0);
		glRotatef(alfa,0,0,1);
		Veiculo.desenhaLinha();
	}
	glPopMatrix();

	if (cont == 1000) {
		if (devoTestar) cout << "Colisao LIGADA.  --->  ";
		else cout << "Colisao DESLIGADA.  --->  ";
		double tempo = (clock() - antes) / (double)CLOCKS_PER_SEC;
		cout << "Tempo gasto nos " << cont << " frames: " << tempo << " segundos" << endl;
		cont = 0;
	}

	//cout << "Antes: " << antes << endl;
	//cout << "Ao final da montagem:" << clock()  << endl;

	glutSwapBuffers();
    // Salva o tempo para o próximo ciclo de rendering
    last_idle_time = time_now;
	//cout << "Apos a exibicao:" << clock()  << endl;
}
void animate2()
{
    glutPostRedisplay();
}
// **********************************************************************
//  void animate ( unsigned char key, int x, int y )
//
// **********************************************************************
void animate()
{
    static float dt;
    static float AccumTime=0;
    
#ifdef _WIN32
    DWORD time_now;
    time_now = GetTickCount();
    dt = (float) (time_now - last_idle_time) / 1000.0;
#else
    // Figure out time elapsed since last call to idle function
    struct timeval time_now;
    gettimeofday(&time_now, NULL);
    dt = (float)(time_now.tv_sec  - last_idle_time.tv_sec) +
    1.0e-6*(time_now.tv_usec - last_idle_time.tv_usec);
#endif
    AccumTime +=dt;
    if (AccumTime >=3) // imprime o FPS a cada 3 segundos
    {
        cout << 1.0/dt << " FPS"<< endl;
        AccumTime = 0;
    }
    // cout << "AccumTime: " << AccumTime << endl;

    // Salva o tempo para o próximo ciclo de rendering
    last_idle_time = time_now;
    
    //if  (GetAsyncKeyState(32) & 0x8000) != 0)
    //  cout << "Espaco Pressionado" << endl;
    
    // Redesenha
    glutPostRedisplay();
}
// **********************************************************************
//  void keyboard ( unsigned char key, int x, int y )
//
// **********************************************************************

void keyboard ( unsigned char key, int x, int y ) {

	switch ( key ) {
	case 27:        // Termina o programa qdo
		exit ( 0 );   // a tecla ESC for pressionada
		break;
	case'r':
		alfa = alfa + 1;
		break;
	case'R':
		alfa = alfa - 1;
		break;
	case' ':
		devoTestar = !devoTestar;
        if (devoTestar) cout << "Colisao LIGADA.  --->  ";
        else cout << "Colisao DESLIGADA.  --->  ";
		break;

	default:
		break;
	}
	glutPostRedisplay();
}


// **********************************************************************
//  void arrow_keys ( int a_keys, int x, int y )
//
//
// **********************************************************************
void arrow_keys ( int a_keys, int x, int y ) {

	float incremento=0.2;

	switch ( a_keys ) {
	case GLUT_KEY_UP:       // Se pressionar UP
		ty += incremento;
		break;
	case GLUT_KEY_DOWN:     // Se pressionar DOWN
		ty -= incremento;
		break;
	case GLUT_KEY_LEFT:       // Se pressionar LEFT
		tx -= incremento;
		break;
	case GLUT_KEY_RIGHT:     // Se pressionar RIGHT
		tx += incremento;
		break;
	default:
		break;
	}
	glutPostRedisplay();
}

// **********************************************************************
//  void main ( int argc, char** argv )
//
//
// **********************************************************************
int  main ( int argc, char** argv ) {
	glutInit            ( &argc, argv );
	glutInitDisplayMode (GLUT_DOUBLE | GLUT_DEPTH | GLUT_RGB );
	glutInitWindowPosition (0,0);

	// Define o tamanho inicial da janela grafica do programa
	glutInitWindowSize  ( 650, 500);

	// Cria a janela na tela, definindo o nome da
	// que aparecera na barra de título da janela.
	glutCreateWindow    ( "Cálculo de colisão." );

	// executa algumas inicializações
	init ();


	// Define que o tratador de evento para
	// o redesenho da tela. A funcao "display"
	// será chamada automaticamente quando
	// for necessário redesenhar a janela
	glutDisplayFunc ( display );
	glutIdleFunc ( animate2 );
	// Define que o tratador de evento para
	// o redimensionamento da janela. A funcao "reshape"
	// será chamada automaticamente quando
	// o usuário alterar o tamanho da janela
	glutReshapeFunc ( reshape );

	// Define que o tratador de evento para
	// as teclas. A funcao "keyboard"
	// será chamada automaticamente sempre
	// o usuário pressionar uma tecla comum
	glutKeyboardFunc ( keyboard );

	// Define que o tratador de evento para
	// as teclas especiais(F1, F2,... ALT-A,
	// ALT-B, Teclas de Seta, ...).
	// A funcao "arrow_keys" será chamada
	// automaticamente sempre o usuário
	// pressionar uma tecla especial
	glutSpecialFunc ( arrow_keys );

	// inicia o tratamento dos eventos
	glutMainLoop ( );

	return 0;
}
